@@ -11,20 +11,14 @@ Released under the MIT License.
 
 @author      Erki Suurjaak
 @created     05.03.2014
-@modified    24.11.2022
+@modified    06.12.2022
 ------------------------------------------------------------------------------
 """
 import collections
-import datetime
-import decimal
-import glob
-import importlib
-import json
+import inspect
 import logging
-import os
-import re
 
-import six
+from . import util
 
 logger = logging.getLogger(__name__)
 
@@ -55,7 +49,7 @@ def init(opts=None, engine=None, **kwargs):
 def fetchall(table, cols="*", where=(), group=(), order=(), limit=(), **kwargs):
     """
     Convenience wrapper for database SELECT and fetch all.
-    Keyword arguments are added to WHERE.
+    Arguments have the same meaning as in select().
     """
     return init().fetchall(table, cols, where, group, order, limit, **kwargs)
 
@@ -63,7 +57,7 @@ def fetchall(table, cols="*", where=(), group=(), order=(), limit=(), **kwargs):
 def fetchone(table, cols="*", where=(), group=(), order=(), limit=(), **kwargs):
     """
     Convenience wrapper for database SELECT and fetch one.
-    Keyword arguments are added to WHERE.
+    Arguments have the same meaning as in select().
     """
     return init().fetchone(table, cols, where, group, order, limit, **kwargs)
 
@@ -72,6 +66,11 @@ def insert(table, values=(), **kwargs):
     """
     Convenience wrapper for database INSERT, returns inserted row ID.
     Keyword arguments are added to VALUES.
+    `table` argument has the same meaning as in select().
+    Column names can be data-class properies as in select().
+
+    @param   values  column values to insert, as dict or a sequence of key-value tuples
+                     or a data object
     """
     return init().insert(table, values, **kwargs)
 
@@ -80,14 +79,23 @@ def select(table, cols="*", where=(), group=(), order=(), limit=(), **kwargs):
     """
     Convenience wrapper for database SELECT, returns database cursor.
 
-    @param   table   table name or expression to select from
+    Column names in `where/cols/group/order` arguments can be data-class properties,
+    like `order=MyClass.my_attribute`.
+
+    @param   table   table/view name or expression to select from,
+                     or a data class with `__name__` used for table name
     @param   where   columns or expressions to select by, as dict or a sequence
-                     of key-value tuples; value can specify operator
+                     of key-value tuples or a data object; value can specify operator
                      e.g. `{"mycolumn": ("IN", [1, 2, 3])}`
     @param   cols    columns to select if not all, as string or a sequence of stringables
+                     or a dict with stringable keys
     @param   group   columns to GROUP BY, as string or a sequence of stringables
+                     or a dict with stringable keys
     @param   order   columns to ORDER BY, as string, a sequence of stringables,
-                     or a combination of column names and direction flags
+                     or a combination of column names and direction flags,
+                     or a dict with column name keys and direction flag values.
+                     Direction flag can be string like `"ASC"`,
+                     or truthy/falsy for ascending/descending.
     @param   limit   LIMIT .. OFFSET.. values, as integer or a sequence of integers;
                      None or -1 disables LIMIT or OFFSET
     @param   kwargs  additional arguments added to WHERE clause
@@ -98,7 +106,11 @@ def select(table, cols="*", where=(), group=(), order=(), limit=(), **kwargs):
 def update(table, values, where=(), **kwargs):
     """
     Convenience wrapper for database UPDATE, returns affected row count.
-    Keyword arguments are added to WHERE.
+    Arguments have the same meaning as in select().
+    Column names can be data-class properies as in select().
+
+    @param   values  column values to set, as dict or a sequence of key-value tuples
+                     or a data object
     """
     return init().update(table, values, where, **kwargs)
 
@@ -106,7 +118,7 @@ def update(table, values, where=(), **kwargs):
 def delete(table, where=(), **kwargs):
     """
     Convenience wrapper for database DELETE, returns affected row count.
-    Keyword arguments are added to WHERE.
+    Arguments have the same meaning as in select().
     """
     return init().delete(table, where, **kwargs)
 
@@ -189,6 +201,20 @@ def register_converter(transformer, typenames, engine=None):
     Engines.get(engine).register_converter(transformer, typenames)
 
 
+def register_row_factory(row_factory, engine=None):
+    """
+    Registers function to produce query results as custom type.
+
+    Registration is global per engine; affects future connections and the current default database.
+
+    @param   row_factory  function(cursor, row tuple) returning row as desired type
+                          or `None` to reset to default. `cursor.description` is a sequence of
+                          7-element tuples, as `(column name, ..engine-specific elements..)`.
+    @param   engine       database engine to register for, defaults to first initialized
+    """
+    Engines.register_row_factory(row_factory, engine)
+
+
 
 class Queryable(object):
     """Abstract base for Database and Transaction."""
@@ -223,10 +249,9 @@ class Queryable(object):
         Convenience wrapper for database SELECT, returns database cursor.
         Keyword arguments are added to WHERE.
         """
-        where = list(where.items() if isinstance(where, dict) else where)
-        where += kwargs.items()
-        sql, args = self.makeSQL("SELECT", table, cols, where, group, order, limit)
-        return self.execute(sql, args)
+        sql, args = self.makeSQL("SELECT", table, cols, where, group, order, limit, kwargs=kwargs)
+        cursor = self.execute(sql, args)
+        return TypeCursor(cursor, table) if inspect.isclass(table) else cursor
 
 
     def update(self, table, values, where=(), **kwargs):
@@ -234,9 +259,7 @@ class Queryable(object):
         Convenience wrapper for database UPDATE, returns affected row count.
         Keyword arguments are added to WHERE.
         """
-        where = list(where.items() if isinstance(where, dict) else where)
-        where += kwargs.items()
-        sql, args = self.makeSQL("UPDATE", table, values=values, where=where)
+        sql, args = self.makeSQL("UPDATE", table, values=values, where=where, kwargs=kwargs)
         return self.execute(sql, args).rowcount
 
 
@@ -245,9 +268,7 @@ class Queryable(object):
         Convenience wrapper for database DELETE, returns affected row count.
         Keyword arguments are added to WHERE.
         """
-        where = list(where.items() if isinstance(where, dict) else where)
-        where += kwargs.items()
-        sql, args = self.makeSQL("DELETE", table, where=where)
+        sql, args = self.makeSQL("DELETE", table, where=where, kwargs=kwargs)
         return self.execute(sql, args).rowcount
 
 
@@ -273,7 +294,8 @@ class Queryable(object):
         raise NotImplementedError()
 
 
-    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(), limit=(), values=()):
+    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(), limit=(), values=(),
+                kwargs=None):
         """Returns (SQL statement string, parameter dict)."""
         raise NotImplementedError()
 
@@ -354,6 +376,23 @@ class Database(Queryable):
         """
         raise NotImplementedError()
 
+    @property
+    def row_factory(self):
+        """The custom row factory, if any, as `function(cursor, row tuple)`."""
+        raise NotImplementedError()
+
+    @row_factory.setter
+    def row_factory(self, row_factory):
+        """
+        Sets custom row factory, as `function(cursor, row tuple)`, or `None` to reset to default.
+
+       `cursor.description` is a sequence of 7-element tuples,
+        first element being column name and the rest engine-specific.
+
+        Overrides globally registered row factory, if any.
+        """
+        raise NotImplementedError()
+
 
 class Transaction(Queryable):
     """
@@ -424,7 +463,7 @@ class Rollback(Exception):
 
 
 # ---------------------------------- detail ----------------------------------
-
+# \cond HIDDENSYMBOLS
 class Engines(object):
     """Database engine broker."""
 
@@ -473,121 +512,59 @@ class Engines(object):
     @classmethod
     def populate(cls):
         """Populates Database engines, if not already populated."""
-        if cls.MODULES is None: cls.MODULES = load_modules()
+        if cls.MODULES is None: cls.MODULES = util.load_modules()
+
+    @classmethod
+    def register_row_factory(cls, row_factory, engine):
+        """Registers row factory for engine, and current engine default database if any."""
+        engine = engine.lower() if engine else next(iter(cls.DATABASES))
+        cls.MODULES[engine].register_row_factory(row_factory)
+        if cls.DATABASES.get(engine): cls.DATABASES[engine].row_factory = row_factory
 
 
-class StaticTzInfo(datetime.tzinfo):
-    """datetime.tzinfo class representing a constant offset from UTC."""
-    ZERO = datetime.timedelta(0)
+class TypeCursor(object):
+    """Wrapper for database cursor, yielding rows constructed with given callable."""
 
-    def __init__(self, name, delta):
-        """Constructs a new static zone info, with specified name and time delta."""
-        self._name    = name
-        self._offset = delta
+    def __init__(self, cursor, callable):
+        """
+        @param   cursor    database engine cursor instance
+        @param   callable  function(rowdict) or function(*row values) or function(**rowdict)
+        """
+        self.__cursor = cursor
+        self.__cls    = callable
+        self.__logged = False
+        for name, value in inspect.getmembers(cursor):
+            if not hasattr(self, name):  # Monkey-patch cursor members to self
+                setattr(self, name, value)
 
-    def utcoffset(self, dt): return self._offset
-    def dst(self, dt):       return self.ZERO
-    def tzname(self, dt):    return self._name
-    def __ne__(self, other): return not self.__eq__(other)
-    def __repr__(self):      return "%s(%s)" % (self.__class__.__name__, self._name)
-    def __eq__(self, other):
-        return isinstance(other, self.__class__) and self._offset == other._offset
-## UTC timezone singleton
-UTC = StaticTzInfo("UTC", StaticTzInfo.ZERO)
-
-
-
-def json_loads(s):
-    """
-    Returns deserialized JSON, with datetime/date strings converted to objects.
-
-    Returns original input if loading as JSON failed.
-    """
-    def convert_recursive(data):
-        """Converts ISO datetime strings to objects in nested dicts or lists."""
+    def fetchmany(self, size=None):
         result = []
-        pairs = enumerate(data) if isinstance(data, list) \
-                else data.items() if isinstance(data, dict) else []
-        rgx = r"^\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}(\.\d+)?(([+-]\d{2}:?\d{2})|Z)?$"
-        for k, v in pairs:
-            if isinstance(v, (dict, list)): v = convert_recursive(v)
-            elif isinstance(v, six.string_types) and len(v) > 18 and re.match(rgx, v):
-                v = parse_datetime(v)
-            result.append((k, v))
-        return [x for _, x in result] if isinstance(data, list) \
-               else dict(result) if isinstance(data, dict) else data
-    try:
-        return None if s is None else json.loads(s, object_hook=convert_recursive)
-    except Exception:
-        fails = getattr(json_loads, "__fails", set())
-        if hash(s) not in fails: # Avoid spamming logs
-            logger.warning("Failed to parse JSON from %r.", s, exc_info=True)
-            setattr(json_loads, "__fails", fails | set([hash(s)]))
-        return s
+        for _ in range(self.__cursor.arraysize if size is None else size):
+            row = next(self.__cursor, None)
+            if row is None: break  # for
+            result.append(self.__factory(row))
+        return result
 
+    def fetchone(self): return next(self, None)
+    def fetchall(self): return list(self)
+    def __iter__(self): return iter(self.__factory(x) for x in self.__cursor)
+    def __next__(self): return self.__factory(next(self.__cursor))
+    def next(self):     return self.__factory(next(self.__cursor))
 
-def json_dumps(data, indent=2, sort_keys=True):
-    """
-    Returns JSON string, with datetime types converted to ISO-8601 strings
-    (in UTC if no timezone set), sets converted to lists,
-    and Decimal objects converted to float or int. Returns None if data is None.
-    """
-    if data is None: return None
-    def encoder(x):
-        if isinstance(x,    set): return list(x)
-        if isinstance(x, (datetime.datetime, datetime.date, datetime.time)):
-            if x.tzinfo is None: x = x.replace(tzinfo=UTC)
-            return x.isoformat()
-        if isinstance(x, decimal.Decimal):
-            return float(x) if x.as_tuple().exponent else int(x)
-        return None
-    return json.dumps(data, default=encoder, indent=indent, sort_keys=sort_keys)
-
-
-def parse_datetime(s):
-    """
-    Tries to parse string as ISO8601 datetime, returns input on error.
-    Supports "YYYY-MM-DD[ T]HH:MM:SS(.micros)?(Z|[+-]HH(:MM)?)?".
-    All returned datetimes are timezone-aware, falling back to UTC.
-    """
-    if len(s) < 18: return s
-    rgx = r"^\d{4}-\d{2}-\d{2}[ T]\d{2}:\d{2}:\d{2}(\.\d+)?(([+-]\d{2}(:?\d{2})?)|Z)?$"
-    result, match = s, re.match(rgx, s)
-    if match:
-        millis, _, offset, _ = match.groups()
-        minimal = re.sub(r"\D", "", s[:match.span(2)[0]] if offset else s)
-        fmt = "%Y%m%d%H%M%S" + ("%f" if millis else "")
-        try:
-            result = datetime.datetime.strptime(minimal, fmt)
-            if offset: # Support timezones like 'Z' or '+03:00'
-                hh, mm = map(int, [offset[1:3], offset[4:]])
-                delta = datetime.timedelta(hours=hh, minutes=mm)
-                if offset.startswith("-"): delta = -delta
-                result = result.replace(tzinfo=StaticTzInfo(offset, delta))
-        except ValueError: pass
-    if isinstance(result, datetime.datetime) and result.tzinfo is None:
-        result = result.replace(tzinfo=UTC) # Force UTC timezone on unaware values
-    return result
-
-
-
-def load_modules():
-    """Returns db engines loaded from file directory, as {name: module}."""
-    result = {}
-    for n in sorted(glob.glob(os.path.join(os.path.dirname(__file__), "engines", "*"))):
-        name = os.path.splitext(os.path.basename(n))[0]
-        if name.startswith("__") or os.path.isfile(n) and not re.match(".*pyc?$", n) \
-        or os.path.isdir(n) and not any(glob.glob(os.path.join(n, x)) for x in ("*.py", "*.pyc")):
-            continue  # for n
-
-        modulename = "%s.%s.%s" % (__package__, "engines", name)
-        module = importlib.import_module(modulename)
-        result[name] = module
-    return result
+    def __factory(self, row):
+        """Returns row constructed with callable, or original row if all argument options failed."""
+        result, errors = util.factory(self.__cls, row)
+        if result is row and not self.__logged:
+            logger.warning("Failed to instantiate %s with keywords, posargs, and dictionary. "
+                           "Returning dictionary.\n%s", self.__cls, "\n".join(map(repr, errors)))
+            self.__logged = True
+        return result
+# \endcond
 
 
 __all__ = [
     "Database", "Rollback", "Transaction",
     "init", "fetchall", "fetchone", "insert", "select", "update", "delete", "execute",
     "executescript", "close", "transaction", "register_adapter", "register_converter",
+    "register_row_factory",
 ]
