@@ -8,13 +8,15 @@ Released under the MIT License.
 
 @author      Erki Suurjaak
 @created     08.05.2020
-@modified    24.11.2022
+@modified    06.12.2022
 ------------------------------------------------------------------------------
 """
 import collections
 from contextlib import contextmanager
+import inspect
 import logging
 import re
+import sys
 import threading
 
 from six.moves import urllib_parse
@@ -27,31 +29,47 @@ try:
     import psycopg2.pool
 except ImportError: psycopg2 = None
 
-from .. import api
+from .. import api, util
 
 logger = logging.getLogger(__name__)
 
 
 ## Postgres reserved keywords, needing quotes in SQL queries
 RESERVED_KEYWORDS = [
-    "ALL", "ANALYSE", "ANALYZE", "AND", "ANY", "ASC", "ASYMMETRIC", "BOTH", "CASE", "CAST", "CHECK",
-    "COLLATE", "COLUMN", "CONSTRAINT", "CURRENT_CATALOG", "CURRENT_DATE", "CURRENT_ROLE",
-    "CURRENT_TIME", "CURRENT_TIMESTAMP", "CURRENT_USER", "DEFAULT", "DEFERRABLE", "DESC",
-    "DISTINCT", "DO", "ELSE", "END", "FALSE", "FOREIGN", "IN", "INITIALLY", "LATERAL", "LEADING",
-    "LOCALTIME", "LOCALTIMESTAMP", "NOT", "NULL", "ONLY", "OR", "PLACING", "PRIMARY", "REFERENCES",
-    "SELECT", "SESSION_USER", "SOME", "SYMMETRIC", "TABLE", "THEN", "TRAILING", "TRUE", "UNIQUE",
-    "USER", "USING", "VARIADIC", "WHEN", "AUTHORIZATION", "BINARY", "COLLATION", "CONCURRENTLY",
-    "CROSS", "CURRENT_SCHEMA", "FREEZE", "FULL", "ILIKE", "INNER", "IS", "JOIN", "LEFT", "LIKE",
-    "NATURAL", "OUTER", "RIGHT", "SIMILAR", "TABLESAMPLE", "VERBOSE", "ISNULL", "NOTNULL",
-    "OVERLAPS", "ARRAY", "AS", "CREATE", "EXCEPT", "FETCH", "FOR", "FROM", "GRANT", "GROUP",
-    "HAVING", "INTERSECT", "INTO", "LIMIT", "OFFSET", "ON", "ORDER", "RETURNING", "TO", "UNION",
-    "WHERE", "WINDOW", "WITH"
+    "ALL", "ANALYSE", "ANALYZE", "AND", "ANY", "ARRAY", "AS", "ASC", "ASYMMETRIC", "AUTHORIZATION",
+    "BINARY", "BOTH", "CASE", "CAST", "CHECK", "COLLATE", "COLLATION", "COLUMN", "CONCURRENTLY",
+    "CONSTRAINT", "CREATE", "CROSS", "CURRENT_CATALOG", "CURRENT_DATE", "CURRENT_ROLE",
+    "CURRENT_SCHEMA", "CURRENT_TIME", "CURRENT_TIMESTAMP", "CURRENT_USER", "DEFAULT", "DEFERRABLE",
+    "DESC", "DISTINCT", "DO", "ELSE", "END", "EXCEPT", "FALSE", "FETCH", "FOR", "FOREIGN", "FREEZE",
+    "FROM", "FULL", "GRANT", "GROUP", "HAVING", "ILIKE", "IN", "INITIALLY", "INNER", "INTERSECT",
+    "INTO", "IS", "ISNULL", "JOIN", "LATERAL", "LEADING", "LEFT", "LIKE", "LIMIT", "LOCALTIME",
+    "LOCALTIMESTAMP", "NATURAL", "NOT", "NOTNULL", "NULL", "OFFSET", "ON", "ONLY", "OR", "ORDER",
+    "OUTER", "OVERLAPS", "PLACING", "PRIMARY", "REFERENCES", "RETURNING", "RIGHT", "SELECT",
+    "SESSION_USER", "SIMILAR", "SOME", "SYMMETRIC", "TABLE", "TABLESAMPLE", "THEN", "TO",
+    "TRAILING", "TRUE", "UNION", "UNIQUE", "USER", "USING", "VARIADIC", "VERBOSE",
+    "WHEN", "WHERE", "WINDOW", "WITH"
 ]
 
 
+class Identifier(object):
+    """Wrapper for table and column names from data objects."""
+    def __init__(self, name): self.name = name
+    def __eq__(self, other):
+        """Supports comparing to strings or other Identifier instances."""
+        if isinstance(other, type(self)): return self.name == other.name
+        return isinstance(other, string_types) and (self.name == other)
+    def __str__(self):
+        """Returns quoted name."""
+        return self.quote(self.name)
+    @staticmethod
+    def quote(name):
+        """Returns quoted name."""
+        return quote(name, force=not name.islower())  # Must quote if not lowercase
+
+
 class Queryable(api.Queryable):
 
-    # Recognized binary operators for makeSQL
+    ## Recognized binary operators for makeSQL()
     OPS = ("!=", "!~", "!~*", "#", "%", "&", "*", "+", "-", "/", "<", "<<",
            "<=", "<>", "<@", "=", ">", ">=", ">>", "@>", "^", "|", "||", "&&", "~",
            "~*", "ANY", "ILIKE", "IN", "IS", "IS NOT", "LIKE", "NOT ILIKE", "NOT IN",
@@ -63,83 +81,114 @@ class Queryable(api.Queryable):
         Convenience wrapper for database INSERT, returns inserted row ID.
         Keyword arguments are added to VALUES.
         """
-        values = list(values.items() if isinstance(values, dict) else values)
-        values += kwargs.items()
-        sql, args = self.makeSQL("INSERT", table, values=values)
+        sql, args = self.makeSQL("INSERT", table, values=values, kwargs=kwargs)
         cursor = self.execute(sql, args)
         row = None if cursor.description is None else next(cursor, None)
         return next(iter(row.values())) if row and isinstance(row, dict) else None
 
 
-    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(), limit=(), values=()):
+    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(), limit=(), values=(),
+                kwargs=None):
         """Returns (SQL statement string, parameter dict)."""
 
         def cast(col, val):
             """Returns column value cast to correct type for use in psycopg."""
-            self._load_schema()
-            field = table in self._structure and self._structure[table]["fields"].get(col)
+            col = col.name if isinstance(col, Identifier) else \
+                  col if isinstance(col, string_types) else \
+                  self._match_name(util.nameify(col, parent=table), tablename)
+            field = tablename in self._structure and self._structure[tablename]["fields"].get(col)
             if field and "array" == field["type"]:
                 return list(listify(val)) # Values for array fields must be lists
             elif field and val is not None:
-                return self._adapt_value(val, field["type"])
+                val = self._adapt_value(val, field["type"])
             if isinstance(val, (list, set)):
                 return tuple(val) # Sequence parameters for IN etc must be tuples
             return val
 
         def parse_members(i, col, op, val):
             """Returns (col, op, val, argkey)."""
+            if isinstance(col, Identifier): col, colsql, pure = col.name, text_type(col), False
+            elif not isinstance(col, string_types):
+                col = self._match_name(util.nameify(col, parent=table), tablename)
+                colsql, pure = Identifier.quote(col), False
+            else: colsql, pure = col, True
             key = "%sW%s" % (re.sub(r"\W+", "_", col), i)
-            if "EXPR" == col.upper():
+            if "EXPR" == col.upper() and pure:
                 # ("EXPR", ("SQL", val))
-                col, op, val, key = val[0], "EXPR", val[1], "EXPRW%s" % i
-            elif col.count("?") == argcount(val):
+                colsql, op, val, key = val[0], "EXPR", val[1], "EXPRW%s" % i
+            elif col.count("?") == argcount(val) and pure:
                 # ("any SQL with ? placeholders", val)
                 op, val, key = "EXPR", listify(val), "EXPRW%s" % i
             elif isinstance(val, (list, tuple)) and len(val) == 2 \
             and isinstance(val[0], string_types):
                 tmp = val[0].strip().upper()
-                if tmp in self.OPS: # ("col", ("binary op like >=", val))
+                if tmp in self.OPS:
+                    # ("col", ("binary op like >=", val))
                     op, val = tmp, val[1]
                 elif val[0].count("?") == argcount(val[1]):
                     # ("col", ("SQL with ? placeholders", val))
-                    col, val, op = "%s = %s" % (col, val[0]), listify(val[1]), "EXPR"
+                    colsql, val, op = "%s = %s" % (col, val[0]), listify(val[1]), "EXPR"
             if op in ("IN", "NOT IN") and not val: # IN -> ANY, to avoid error on empty array
-                col = "%s%s = ANY('{}')" % ("" if "IN" == op else "NOT ", col)
+                colsql = "%s%s = ANY('{}')" % ("" if "IN" == op else "NOT ", colsql)
                 op = "EXPR"
-            return col, op, val, key
+            return colsql, op, cast(col, val), key
         def argcount(x)  : return len(x) if isinstance(x, (list, set, tuple)) else 1
-        def listify(x)   : return x if isinstance(x, (list, tuple)) else [x]
-        def strlistify(x): return [text_type(y) for y in listify(x) if y not in ("", None)]
+        def listify(x)   : return x if isinstance(x, (list, tuple)) else \
+                                  list(x) if isinstance(x, set) else [x]
+        def keylistify(x): return x if isinstance(x, (list, tuple)) else \
+                                  list(x) if isinstance(x, (dict, set)) else [x]
 
+        def column(val, sql=False):
+            """Returns column name from string/property/Identifier, quoted if object and `sql`."""
+            if inspect.isdatadescriptor(val): val = util.nameify(val, wrapper(sql=sql), table)
+            if isinstance(val, Identifier): return text_type(val) if sql else val.name
+            return val if isinstance(val, string_types) else text_type(val)
+
+        def wrapper(column=True, sql=False):
+            """Returns function producing Identifier or SQL-ready name string."""
+            def inner(name):
+                val = Identifier(self._match_name(name, table=tablename if column else None))
+                return text_type(val) if sql else val
+            return inner
+
+        self._load_schema()
+        values0 = values
         action = action.upper()
-        cols   =    cols if isinstance(cols,  string_types) else ", ".join(strlistify(cols)) or "*"
-        where  = [where] if isinstance(where, string_types) else where
-        group  =   group if isinstance(group, string_types) else ", ".join(strlistify(group))
-        order  = [order] if isinstance(order, string_types) else order
+        where, group, order, limit, values = (() if x is None else x
+                                              for x in (where, group, order, limit, values))
+        n = util.nameify(table, wrapper(column=False))
+        tablename, tablesql = (n.name, text_type(n)) if isinstance(n, Identifier) else (n, n)
+        cols   = ", ".join(util.nameify(x, wrapper(sql=True), table) for x in keylistify(cols)) or "*"
+        group  = ", ".join(util.nameify(x, wrapper(sql=True), table) for x in keylistify(group))
+        where  = util.keyvalues(where, wrapper())
+        order  = list(order.items()) if isinstance(order, dict) else listify(order)
         order  = [order] if isinstance(order, (list, tuple)) \
                  and len(order) == 2 and isinstance(order[1], bool) else order
         limit  = [limit] if isinstance(limit, string_types + integer_types) else limit
-        values = values if not isinstance(values, dict) else values.items()
-        where  =  where if not isinstance(where,  dict)  else where.items()
-        sql = "SELECT %s FROM %s" % (cols, table) if "SELECT" == action else ""
-        sql = "DELETE FROM %s"    % (table)       if "DELETE" == action else sql
-        sql = "INSERT INTO %s"    % (table)       if "INSERT" == action else sql
-        sql = "UPDATE %s"         % (table)       if "UPDATE" == action else sql
-        args = {}
+        values = util.keyvalues(values, wrapper())
+        sql    = "SELECT %s FROM %s" % (cols, tablesql) if "SELECT" == action else ""
+        sql    = "DELETE FROM %s"    % (tablesql)       if "DELETE" == action else sql
+        sql    = "INSERT INTO %s"    % (tablesql)       if "INSERT" == action else sql
+        sql    = "UPDATE %s"         % (tablesql)       if "UPDATE" == action else sql
+        args   = {}
+        if kwargs and action in ("SELECT", "DELETE", "UPDATE"): where  += list(kwargs.items())
+        if kwargs and action in ("INSERT", ):                   values += list(kwargs.items())
 
         if "INSERT" == action:
-            keys = ["%sI%s" % (re.sub(r"\W+", "_", k), i) for i, (k, _) in enumerate(values)]
-            args.update((n, cast(k, v)) for n, (k, v) in zip(keys, values))
-            cols, vals = ", ".join(k for k, _ in values), ", ".join("%%(%s)s" % n for n in keys)
+            pk = self._structure.get(tablename, {}).get("key")
+            if pk and util.is_dataobject(values0):  # Can't avoid giving primary key if data object
+                values = [(k, v) for k, v in values if k != pk or v is not None]  # Discard NULL pk
+            keys = ["%sI%s" % (re.sub(r"\W+", "_", column(k)), i) for i, (k, _) in enumerate(values)]
+            args.update((a, cast(k, v)) for i, (a, (k, v)) in enumerate(zip(keys, values)))
+            cols = ", ".join(column(k, sql=True) for k, _ in values)
+            vals = ", ".join("%%(%s)s" % s for s in keys)
             sql += " (%s) VALUES (%s)" % (cols, vals)
-            self._load_schema()
-            if table in self._structure and self._structure[table].get("key"):
-                sql += " RETURNING %s AS id" % (self._structure[table]["key"])
+            if pk: sql += " RETURNING %s AS id" % Identifier.quote(pk)
         if "UPDATE" == action:
             sql += " SET "
             for i, (col, val) in enumerate(values):
-                key = "%sU%s" % (re.sub(r"\W+", "_", col), i)
-                sql += (", " if i else "") + "%s = %%(%s)s" % (col, key)
+                key = "%sU%s" % (re.sub(r"\W+", "_", column(col)), i)
+                sql += (", " if i else "") + "%s = %%(%s)s" % (column(col, sql=True), key)
                 args[key] = cast(col, val)
         if where:
             sql += " WHERE "
@@ -163,18 +212,17 @@ class Queryable(api.Queryable):
                     op = {"=": "IS", "!=": "IS NOT", "<>": "IS NOT"}.get(op, op)
                     sql += (" AND " if i else "") + "%s %s NULL" % (col, op)
                 else:
-                    args[key] = cast(col, val)
+                    args[key] = val
                     sql += (" AND " if i else "") + "%s %s %%(%s)s" % (col, op, key)
         if group:
             sql += " GROUP BY " + group
         if order:
             sql += " ORDER BY "
-            for i, col in enumerate(listify(order)):
-                name = col if isinstance(col, string_types) else \
-                       text_type(col[0] if isinstance(col, (list, tuple)) else col)
+            for i, col in enumerate(order):
+                name = util.nameify(col[0] if isinstance(col, (list, tuple)) else col, quote, table)
                 sort = col[1] if name != col and isinstance(col, (list, tuple)) and len(col) > 1 \
                        else ""
-                if not isinstance(sort, string_types): sort = "DESC" if sort else ""
+                if not isinstance(sort, string_types): sort = "" if sort else "DESC"
                 sql += (", " if i else "") + name + (" " if sort else "") + sort
         if limit:
             limit = [None if isinstance(v, integer_types) and v < 0 else v for v in limit]
@@ -183,6 +231,7 @@ class Queryable(api.Queryable):
                 sql += " %s %%(%s)s" % (k.upper(), k)
                 args[k] = v
 
+        logger.log(logging.DEBUG // 2, sql)
         return sql, args
 
 
@@ -204,7 +253,7 @@ class Queryable(api.Queryable):
         or original value.
         """
         if typename in ("json", "jsonb") and type(value) not in self.ADAPTERS.values():
-            return psycopg2.extras.Json(value, dumps=api.json_dumps)
+            return psycopg2.extras.Json(value, dumps=util.json_dumps)
         return value
 
 
@@ -212,8 +261,29 @@ class Queryable(api.Queryable):
         """Populates table structure from database if uninitialized or forced."""
         if self._structure is None or force:
             self._structure = {}  # Avoid recursion on first query
-            self._structure.update(query_schema(self, keys=True))
 
+            self.cursor.factory, factory0 = None, self.cursor.factory  # Ensure dict rows
+            try: self._structure.update(query_schema(self, keys=True))
+            finally: self.cursor.factory = factory0
+
+
+    def _match_name(self, name, table=None):
+        """
+        Returns proper-cased name from schema lookup, or same value if no match.
+
+        @parma   name     name of table/view or field, from data object or property
+        @param   table    name of table to match field for, if not matching table name
+        """
+        container = self._structure.get(table, {}).get("fields") if table else self._structure
+        if name not in (container or {}):  # Check for case differences
+            namelc = name.lower()
+            if namelc in container:        # Normal lower-case name present
+                name = namelc
+            elif namelc == name:           # Name from data is lowercase, check for single cased
+                variants = [n for n in container if n.lower() == namelc]
+                if len(variants) == 1:
+                    name = variants[0]
+        return name
 
 
 class Database(api.Database, Queryable):
@@ -238,6 +308,9 @@ class Database(api.Database, Queryable):
     ## Connection pools, as {Database: psycopg2.pool.ConnectionPool}
     POOLS = {}
 
+    ## Registered row factory
+    ROW_FACTORY = None
+
 
     def __init__(self, opts, **kwargs):
         """
@@ -255,12 +328,13 @@ class Database(api.Database, Queryable):
         """
 
         ## Data Source Name, as URL like `"postgresql://user@host/dbname"`
-        self.dsn        = make_db_url(opts)
-        self._kwargs    = kwargs
-        self._cursor    = None
-        self._cursorctx = None
-        self._txs       = []  # [Transaction, ]
-        self._structure = None  # Database schema as {table or view name: {"fields": {..}, ..}}
+        self.dsn          = make_db_url(opts)
+        self._kwargs      = kwargs
+        self._cursor      = None
+        self._cursorctx   = None
+        self._txs         = []    # [Transaction, ]
+        self._row_factory = None  # None if default, False if explicitly default, or func(cur, row)
+        self._structure   = None  # Database schema as {table or view name: {"fields": {..}, ..}}
 
 
     def __enter__(self):
@@ -343,6 +417,24 @@ class Database(api.Database, Queryable):
         return self._cursor
 
 
+    @property
+    def row_factory(self):
+        """The custom row factory, if any, as `function(cursor, row tuple)`."""
+        return None if self._row_factory in (False, None) else self._row_factory
+
+
+    @row_factory.setter
+    def row_factory(self, row_factory):
+        """
+        Sets custom row factory, as `function(cursor, row tuple)`, or `None` to reset to default.
+
+        `cursor.description` is a sequence of 7-element tuples,
+        as `(name, type_code, display_size, internal_size, precision, scale, null_ok)`.
+        """
+        self._row_factory = False if row_factory is None else row_factory
+        if self._cursor: self._cursor.factory = row_factory
+
+
     def transaction(self, commit=True, exclusive=False, **kwargs):
         """
         Returns a transaction context manager.
@@ -372,7 +464,7 @@ class Database(api.Database, Queryable):
         @param   lazy        if true, returns a named server-side cursor that fetches rows
                              iteratively in batches; only supports making a single query
         @param   itersize    batch size in rows for server-side cursor
-        @return              psycopg2.extras.RealDictCursor
+        @return              psycopg2 Cursor
         """
         connection = self.POOLS[self].getconn()
         try:
@@ -413,7 +505,7 @@ class Database(api.Database, Queryable):
             if db in cls.POOLS: return
 
             args = minconn, maxconn, db.dsn
-            kwargs.update(cursor_factory=psycopg2.extras.RealDictCursor)
+            kwargs.update(cursor_factory=db._cursor_factory)
             cls.POOLS[db] = psycopg2.pool.ThreadedConnectionPool(*args, **kwargs)
 
 
@@ -431,6 +523,13 @@ class Database(api.Database, Queryable):
                 psycopg2.extensions.register_type(TYPE)
 
 
+    def _cursor_factory(self, *args, **kwargs):
+        """Returns a new RowFactoryCursor."""
+        factory = self._row_factory
+        if factory in (False, None): factory = None if factory is False else self.ROW_FACTORY
+        return RowFactoryCursor(factory, *args, **kwargs)
+
+
     def _notify(self, tx):
         """Notifies database of transaction closing."""
         if tx in self._txs: self._txs.remove(tx)
@@ -582,6 +681,37 @@ class Transaction(api.Transaction, Queryable):
         return super(Transaction, self)._load_schema(force=force)
 
 
+
+class RowFactoryCursor(psycopg2.extensions.cursor if psycopg2 else object):
+    """A cursor that generates result rows via given factory callable."""
+
+    def __init__(self, row_factory, *args, **kwargs):
+        self.factory = row_factory
+        self.rowtype = dict if sys.version_info > (3, ) else collections.OrderedDict
+        super(RowFactoryCursor, self).__init__(*args, **kwargs)
+
+    def fetchone(self):
+        row = super(RowFactoryCursor, self).fetchone()
+        return row if row is None else self.row_factory(row)
+
+    def fetchmany(self, size=None):
+        rows = super(RowFactoryCursor, self).fetchmany(size)
+        return [self.row_factory(row) for row in rows]
+
+    def fetchall(self):
+        rows = super(RowFactoryCursor, self).fetchall()
+        return [self.row_factory(row) for row in rows]
+
+    def __next__(self): return self.row_factory(super(RowFactoryCursor, self).__next__())
+    def next(self):     return self.__next__()
+
+    def row_factory(self, row):
+        """Returns value constructed with custom row factory, or as dict if no factory."""
+        if self.factory not in (False, None): return self.factory(self, row)
+        return self.rowtype(zip([x[0] for x in self.description], row))
+
+
+
 def autodetect(opts):
     """
     Returns true if input is recognizable as Postgres connection options.
@@ -744,6 +874,11 @@ def register_converter(transformer, typenames):
     Database.CONVERTERS.update({n: transformer for n in typenames if n not in ("JSON", "JSONB")})
 
 
+def register_row_factory(row_factory):
+    """Registers custom row factory, as or `None` to reset to default."""
+    Database.ROW_FACTORY = row_factory
+
+
 
 if psycopg2:
     try:
@@ -754,5 +889,5 @@ if psycopg2:
 
 __all__ = [
     "RESERVED_KEYWORDS", "Database", "Transaction",
-    "autodetect", "quote", "register_adapter", "register_converter",
+    "autodetect", "quote", "register_adapter", "register_converter", "register_row_factory",
 ]
