@@ -8,10 +8,11 @@ Released under the MIT License.
 
 @author      Erki Suurjaak
 @created     05.03.2014
-@modified    25.11.2022
+@modified    06.12.2022
 ------------------------------------------------------------------------------
 """
 import collections
+import inspect
 import logging
 import os
 import re
@@ -21,7 +22,7 @@ import threading
 
 from six import binary_type, integer_types, string_types, text_type
 
-from .. import api
+from .. import api, util
 
 logger = logging.getLogger(__name__)
 
@@ -45,6 +46,7 @@ RESERVED_KEYWORDS = [
 
 class Queryable(api.Queryable):
 
+    ## Recognized binary operators for makeSQL()
     OPS = ["||", "*", "/", "%", "+", "-", "<<", ">>", "&", "|", "<", "<=", ">",
            ">=", "=", "==", "!=", "<>", "IS", "IS NOT", "IN", "NOT IN", "LIKE",
            "GLOB", "MATCH", "REGEXP", "AND", "OR"]
@@ -55,13 +57,12 @@ class Queryable(api.Queryable):
         Convenience wrapper for database INSERT, returns inserted row ID.
         Keyword arguments are added to VALUES.
         """
-        values = list(values.items() if isinstance(values, dict) else values)
-        values += kwargs.items()
-        sql, args = self.makeSQL("INSERT", table, values=values)
+        sql, args = self.makeSQL("INSERT", table, values=values, kwargs=kwargs)
         return self.execute(sql, args).lastrowid
 
 
-    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(), limit=(), values=()):
+    def makeSQL(self, action, table, cols="*", where=(), group=(), order=(), limit=(), values=(),
+                kwargs=None):
         """Returns (SQL statement string, parameter dict)."""
 
         def cast(col, val):
@@ -70,6 +71,7 @@ class Queryable(api.Queryable):
 
         def parse_members(i, col, op, val):
             """Returns (col, op, val, argkey)."""
+            col = util.nameify(col, quote, table)
             key = "%sW%s" % (re.sub(r"\W+", "_", col), i)
             if "EXPR" == col.upper():
                 # ("EXPR", ("SQL", val))
@@ -80,42 +82,55 @@ class Queryable(api.Queryable):
             elif isinstance(val, (list, tuple)) and len(val) == 2 \
             and isinstance(val[0], string_types):
                 tmp = val[0].strip().upper()
-                if tmp in self.OPS: # ("col", ("binary op like >=", val))
+                if tmp in self.OPS:
+                    # ("col", ("binary op like >=", val))
                     op, val = tmp, val[1]
                 elif val[0].count("?") == argcount(val[1]):
                     # ("col", ("SQL with ? placeholders", val))
                     col, val, op = "%s = %s" % (col, val[0]), listify(val[1]), "EXPR"
             return col, op, val, key
         def argcount(x)  : return len(x) if isinstance(x, (list, set, tuple)) else 1
-        def listify(x)   : return x if isinstance(x, (list, tuple)) else [x]
-        def strlistify(x): return [text_type(y) for y in listify(x) if y not in ("", None)]
+        def listify(x)   : return x if isinstance(x, (list, tuple)) else \
+                                  list(x) if isinstance(x, set) else [x]
+        def keylistify(x): return x if isinstance(x, (list, tuple)) else \
+                                  list(x) if isinstance(x, (dict, set)) else [x]
+
+        def column(val, sql=False):
+            """Returns column name from string/property, quoted if object and `sql`."""
+            if inspect.isdatadescriptor(val): val = util.nameify(val, quote if sql else None, table)
+            return val if isinstance(val, string_types) else text_type(val)
 
         action = action.upper()
-        cols   =    cols if isinstance(cols,  string_types) else ", ".join(strlistify(cols)) or "*"
-        where  = [where] if isinstance(where, string_types) else where
-        group  =   group if isinstance(group, string_types) else ", ".join(strlistify(group))
-        order  = [order] if isinstance(order, string_types) else order
+        where, group, order, limit, values = (() if x is None else x
+                                              for x in (where, group, order, limit, values))
+        tablesql = util.nameify(table, quote)
+        cols   = ", ".join(util.nameify(x, quote, table) for x in keylistify(cols)) or "*"
+        group  = ", ".join(util.nameify(x, quote, table) for x in keylistify(group))
+        where  = util.keyvalues(where, quote)
+        order  = list(order.items()) if isinstance(order, dict) else listify(order)
         order  = [order] if isinstance(order, (list, tuple)) \
                  and len(order) == 2 and isinstance(order[1], bool) else order
         limit  = [limit] if isinstance(limit, string_types + integer_types) else limit
-        values = values if not isinstance(values, dict) else values.items()
-        where  =  where if not isinstance(where,  dict)  else where.items()
-        sql = "SELECT %s FROM %s" % (cols, table) if "SELECT" == action else ""
-        sql = "DELETE FROM %s"    % (table)       if "DELETE" == action else sql
-        sql = "INSERT INTO %s"    % (table)       if "INSERT" == action else sql
-        sql = "UPDATE %s"         % (table)       if "UPDATE" == action else sql
-        args = {}
+        values = util.keyvalues(values, quote)
+        sql    = "SELECT %s FROM %s" % (cols, tablesql) if "SELECT" == action else ""
+        sql    = "DELETE FROM %s"    % (tablesql)       if "DELETE" == action else sql
+        sql    = "INSERT INTO %s"    % (tablesql)       if "INSERT" == action else sql
+        sql    = "UPDATE %s"         % (tablesql)       if "UPDATE" == action else sql
+        args   = {}
+        if kwargs and action in ("SELECT", "DELETE", "UPDATE"): where  += list(kwargs.items())
+        if kwargs and action in ("INSERT", ):                   values += list(kwargs.items())
 
         if "INSERT" == action:
-            keys = ["%sI%s" % (re.sub(r"\W+", "_", k), i) for i, (k, _) in enumerate(values)]
+            keys = ["%sI%s" % (re.sub(r"\W+", "_", column(k)), i) for i, (k, _) in enumerate(values)]
             args.update((n, cast(k, v)) for n, (k, v) in zip(keys, values))
-            cols, vals = ", ".join(k for k, _ in values), ", ".join(":%s" % n for n in keys)
+            cols = ", ".join(column(k, sql=True) for k, _ in values)
+            vals = ", ".join(":%s" % n for n in keys)
             sql += " (%s) VALUES (%s)" % (cols, vals)
         if "UPDATE" == action:
             sql += " SET "
             for i, (col, val) in enumerate(values):
-                key = "%sU%s" % (re.sub(r"\W+", "_", col), i)
-                sql += (", " if i else "") + "%s = :%s" % (col, key)
+                key = "%sU%s" % (re.sub(r"\W+", "_", column(col)), i)
+                sql += (", " if i else "") + "%s = :%s" % (column(col, sql=True), key)
                 args[key] = cast(col, val)
         if where:
             sql += " WHERE "
@@ -150,12 +165,11 @@ class Queryable(api.Queryable):
             sql += " GROUP BY " + group
         if order:
             sql += " ORDER BY "
-            for i, col in enumerate(listify(order)):
-                name = col if isinstance(col, string_types) else \
-                       text_type(col[0] if isinstance(col, (list, tuple)) else col)
+            for i, col in enumerate(order):
+                name = util.nameify(col[0] if isinstance(col, (list, tuple)) else col, quote, table)
                 sort = col[1] if name != col and isinstance(col, (list, tuple)) and len(col) > 1 \
                        else ""
-                if not isinstance(sort, string_types): sort = "DESC" if sort else ""
+                if not isinstance(sort, string_types): sort = "" if sort else "DESC"
                 sql += (", " if i else "") + name + (" " if sort else "") + sort
         if limit:
             limit = [None if isinstance(v, integer_types) and v < 0 else v for v in limit]
@@ -166,9 +180,11 @@ class Queryable(api.Queryable):
                 sql += " %s :%s" % (k.upper(), k)
                 args[k] = v
 
+        logger.log(logging.DEBUG // 2, sql)
         return sql, args
 
 
+
     @classmethod
     def quote(cls, value, force=False):
         """
@@ -192,6 +208,9 @@ class Database(api.Database, Queryable):
     ## Mutexes for exclusive actions, as {Database instance: lock}
     MUTEX = collections.defaultdict(threading.RLock)
 
+    ## Registered row factory
+    ROW_FACTORY = None
+
 
     def __init__(self, opts=":memory:", **kwargs):
         """
@@ -202,10 +221,14 @@ class Database(api.Database, Queryable):
                          like `detect_types=sqlite3.PARSE_COLNAMES`
         """
         super(Database, self).__init__()
-        self.connection = None
-        self.path       = opts
-        self._kwargs    = kwargs
-        self._txs       = []  # [Transaction, ]
+        self.connection   = None
+        self.path         = opts
+        self._kwargs      = kwargs
+        self._isolevel    = None  # Connection isolation level, None is auto-commit
+        rowtype = dict if sys.version_info > (3, ) else collections.OrderedDict
+        self._def_factory = lambda cursor, row: rowtype(sqlite3.Row(cursor, row))
+        self._row_factory = None  # None if default, False if explicitly default, or func(cur, row)
+        self._txs         = []    # [Transaction, ]
 
 
     def __enter__(self):
@@ -253,9 +276,11 @@ class Database(api.Database, Queryable):
         if ":memory:" != self.path and not os.path.exists(self.path):
             try: os.makedirs(os.path.dirname(self.path))
             except Exception: pass
-        conn = sqlite3.connect(self.path, **args)
-        conn.row_factory = lambda cursor, row: dict(sqlite3.Row(cursor, row))
-        self.connection = conn
+        self.connection = sqlite3.connect(self.path, **args)
+        self._isolevel = self.connection.isolation_level
+        row_factory = self.ROW_FACTORY if self._row_factory is None else self._row_factory
+        if row_factory in (False, None): row_factory = self._def_factory
+        self.connection.row_factory = row_factory
 
 
     def close(self, commit=None):
@@ -284,6 +309,25 @@ class Database(api.Database, Queryable):
         return self.connection.cursor() if self.connection else None
 
 
+    @property
+    def row_factory(self):
+        """The custom row factory, if any, as `function(cursor, row tuple)`."""
+        return None if self._row_factory in (False, None) else self._row_factory
+
+
+    @row_factory.setter
+    def row_factory(self, row_factory):
+        """
+        Sets custom row factory, as `function(cursor, row tuple)`, or `None` to reset to default.
+
+       `cursor.description` is a sequence of 7-element tuples, as `(column name, None, None, ..)`.
+        """
+        self._row_factory = False if row_factory is None else row_factory
+        if self.connection:
+            factory = self._def_factory if self._row_factory is False else self._row_factory
+            self.connection.row_factory = factory
+
+
     def transaction(self, commit=True, exclusive=True, **kwargs):
         """
         Returns a transaction context manager.
@@ -303,6 +347,7 @@ class Database(api.Database, Queryable):
     def _notify(self, tx):
         """Notifies database of transaction closing."""
         if tx in self._txs: self._txs.remove(tx)
+        if not self._txs and self.connection: self.connection.isolation_level = self._isolevel
 
 
 
@@ -327,7 +372,6 @@ class Transaction(api.Transaction, Queryable):
         self._db         = db
         self._exitcommit = commit
         self._enterstack = 0  # Number of levels the transaction context is nested at
-        self._isolevel0  = db.connection.isolation_level
         self._exclusive  = True if exclusive is None else exclusive
         self._closed     = False
         self._cursor     = None
@@ -355,7 +399,6 @@ class Transaction(api.Transaction, Queryable):
             if depth < 1:
                 self._cursor = None
                 self._closed = True
-                self._db.connection.isolation_level = self._isolevel0
                 self._db._notify(self)
             if self._exclusive: Database.MUTEX[self._db].release()
 
@@ -373,7 +416,6 @@ class Transaction(api.Transaction, Queryable):
         elif commit or self._exitcommit: self.commit()
         self._cursor = None
         self._closed = True
-        self._db.connection.isolation_level = self._isolevel0
         self._db._notify(self)
 
     def execute(self, sql, args=()):
@@ -490,7 +532,12 @@ def register_converter(transformer, typenames):
     for n in typenames: sqlite3.register_converter(n, transformer)
 
 
+def register_row_factory(row_factory):
+    """Registers custom row factory, as or `None` to reset to default."""
+    Database.ROW_FACTORY = row_factory
+
+
 __all__ = [
     "RESERVED_KEYWORDS", "Database", "Transaction",
-    "autodetect", "quote", "register_adapter", "register_converter",
+    "autodetect", "quote", "register_adapter", "register_converter", "register_row_factory",
 ]
