@@ -11,14 +11,14 @@ Released under the MIT License.
 
 @author      Erki Suurjaak
 @created     22.11.2022
-@modified    22.11.2022
+@modified    06.12.2022
 ------------------------------------------------------------------------------
 """
 import collections
+import contextlib
 import datetime
 import json
 import logging
-import os
 import unittest
 
 import dblite
@@ -45,9 +45,9 @@ class TestTransformers(unittest.TestCase):
     ## Table test data, as {table name: [{row}]}
     DATAS = {
         "test": [
-            {"id": 1, "dt": datetime.datetime.now(dblite.api.UTC),
+            {"id": 1, "dt": datetime.datetime.now(dblite.util.UTC),
              "val": {"nested": {"value": [1, 2]}}},
-            {"id": 2, "dt": datetime.datetime.now(dblite.api.UTC),
+            {"id": 2, "dt": datetime.datetime.now(dblite.util.UTC),
              "val": {"nested": [None, False, 1.1, 2.2]}},
         ],
     }
@@ -71,7 +71,7 @@ class TestTransformers(unittest.TestCase):
             return
         opts, kwargs = self.ENGINES["postgres"]
         try:
-            dblite.init(opts, **kwargs)
+            dblite.init(opts, **kwargs).close()
         except psycopg2.Error as e:
             logger.warning("Skip testing postgres, connection failed with:\n%s", e)
         else:
@@ -80,8 +80,11 @@ class TestTransformers(unittest.TestCase):
 
 
     def tearDown(self):
-        """Deletes temoorary files."""
-        try: os.remove(self._path)
+        """Drops created tables from Postgres."""
+        try:
+            opts, kwargs = self._connections["postgres"]
+            with dblite.init(opts, "postgres", **kwargs) as db:
+                for table in self.TABLES: db.executescript("DROP TABLE IF EXISTS %s" % table)
         except Exception: pass
         super(TestTransformers, self).tearDown()
 
@@ -89,13 +92,15 @@ class TestTransformers(unittest.TestCase):
     def test_transformers(self):
         """Tests adapters and converters."""
         logger.info("Verifying transformer functions.")
-        for engine, (opts, kwargs) in self._connections.items():
-            dblite.init(opts, **kwargs)
-            dblite.register_adapter(json.dumps, dict)
-            dblite.register_converter(json.loads, "JSON")
-            self.verify_transformers(engine)
-            dblite.close()
+        for i, (engine, (opts, kwargs)) in enumerate(self._connections.items()):
             dblite.api.Engines.DATABASES.clear()  # Clear cache of default databases
+            with self.subTest(engine) if hasattr(self, "subTest") else contextlib.nested():  # Py3/2
+                if i: logger.info("-" * 60)
+                dblite.init(opts, **kwargs)
+                dblite.register_adapter(json.dumps, dict)
+                dblite.register_converter(json.loads, "JSON")
+                self.verify_transformers(engine)
+                dblite.close()
 
 
     def verify_transformers(self, engine):
@@ -115,6 +120,8 @@ class TestTransformers(unittest.TestCase):
                 row["dt"], data["dt"] = (x["dt"].replace(tzinfo=None) for x in (row, data))
                 self.assertEqual(row, data, "Unexpected value from dblite.select().")
 
+            dblite.executescript("DROP TABLE %s" % table)
+
 
 
 if "__main__" == __name__:
